## HEADER ####
# Who: Ed Harris
# Last edited: 2020.05.25
# What: Deep learning with Keras and Tensorflow first dive
## Discussion with the Global Wheat Challenge Kaggle group
## Can we and should we do deep learning using R?
####

## CONTENTS ####
## 00 Setup
## 01 Preparing data

## 00 Setup ####
# Following this excellent tutorial...
# https://tensorflow.rstudio.com/guide/keras/


# install.packages("keras", dep = T)
# install.packages("tensorflow", dep = T)
# install.packages("reticulate", dep=T)

library(keras)      # Interface to Keras, allows R to talk to Keras, high level
library(tensorflow) # Interface to Tensorflow, low level, allows GPU utilisation
library(reticulate) # Interface to Python libraries etc ... 
library(tictoc)

## 01 Preparing data ####
# https://tensorflow.rstudio.com/guide/keras/#preparing-the-data
# PREPARING THE DATA

# Dataset of loads of handwritten integers
# Each is a 28 x 28 grayscale picture
# Idea is to use classification to identify the integer
mnist <- dataset_mnist() #NB Cuda message

# Training versus testing ####
# x is the image data
# y is the vector that identifies the integer
x_train <- mnist$train$x
y_train <- mnist$train$y
x_test <- mnist$test$x
y_test <- mnist$test$y

# look at data structure
class(x_train)
dim(x_train)

image <- 8
x_train[image,,]

image(t(x_train[image,,]), 
      useRaster=TRUE, axes=FALSE,
      ylim = c(1,0), xlim = c(0,1),
      main = y_train[image],
      col=gray.colors(254))


# reshape ####
# shape the image data into vectors (NB all the same shape)
x_train <- array_reshape(x_train, c(nrow(x_train), 784))
x_test <- array_reshape(x_test, c(nrow(x_test), 784))

class(x_train)
dim(x_train)

x_train[image, ]

# rescale ####

x_train <- x_train / 255
x_test <- x_test / 255

x_train[image, ]


# keras one-hot encode ####
# https://www.quora.com/What-is-one-hot-encoding-and-when-is-it-used-in-data-science

# "one hot encoding" codes complex data into binary data == categorical
# faster for machine classification...

# from this...
y_train[1:50] # vector of "solutions"
y_train[1:2]; dim(y_train) # note 5 in position 1, zero position 2

y_train <- to_categorical(y = y_train, num_classes = 10) # note 5 in position 1, zero position 2
y_test <- to_categorical(y = y_test, num_classes = 10)

# to this
y_train[1:2,]; dim(y_train)

## https://tensorflow.rstudio.com/guide/keras/#defining-the-model
## DEFINING THE MODEL

## Sequntial model
## https://keras.rstudio.com/articles/sequential_model.html

# CPU versus GPU https://tensorflow.rstudio.com/installation/gpu/local_gpu/
# https://www.tensorflow.org/install/gpu#hardware_requirements

## DEFINE THE MODEL ####
# This is an important step
# "playing" with the inputs here is an important step in machine learning...

# There are many different models
# this is the sequential model...

model <- keras_model_sequential() 
model %>% 
  layer_dense(units = 256, activation = 'relu', input_shape = c(784)) %>% 
  layer_dropout(rate = 0.4) %>% 
  layer_dense(units = 128, activation = 'relu') %>%
  layer_dropout(rate = 0.3) %>%
  layer_dense(units = 10, activation = 'softmax')

summary(model) # this is info about the model setup and initial parameters
# input_shape = c(784)

## COMPILE THE MODEL ####
model %>% compile(
  loss = 'categorical_crossentropy',
  optimizer = optimizer_rmsprop(),
  metrics = c('accuracy')
)

## TRAIN AND EVALUATE THE MODEL ####
tic()
history <- model %>% fit(
  x_train, y_train, 
  epochs = 30, batch_size = 128, 
  validation_split = 0.2
)
toc()

plot(history)

## EVALUATE MODEL ACCURACY ####
# (show me the money)
model %>% evaluate(x_test, y_test)

## Generate predictions on new data ####
model %>% predict_classes(x_test) 

## look closer ####
testset <- 11:15
y_test[testset,]
model %>% predict_classes(x_test[testset,]) #not bad!
